"""Unit tests for Generalised Pareto Distribution (GPD)."""

import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))

import pytest
import numpy as np

from pyreto import gpd


def generate_extreme_returns(seed=42, size=500):
    """Generate extreme returns suitable for POT/GPD fitting."""
    np.random.seed(seed)
    # Mix of normal returns and extreme values
    normal_returns = np.random.normal(0, 0.01, size)

    # Add some extreme values (exceedances)
    extreme_prob = 0.05  # 5% extreme events
    extreme_mask = np.random.random(size) < extreme_prob
    extreme_values = -0.03 - np.random.exponential(0.02, size)  # Negative extremes

    returns = normal_returns.copy()
    returns[extreme_mask] = extreme_values[extreme_mask]

    return returns


class TestGPDFit:
    """Test GPD fitting functionality."""

    def test_fit_basic(self):
        """Test basic GPD fitting."""
        returns = generate_extreme_returns(seed=1, size=500)
        par = gpd.fit(returns)

        assert isinstance(par, dict), "Should return dict"
        assert 'c' in par and 'loc' in par and 'scale' in par
        assert par['scale'] > 0, f"Scale should be positive, got {par['scale']}"

    def test_fit_with_heavy_tails(self):
        """Test fitting with heavy-tailed data."""
        np.random.seed(123)
        # Generate data with heavy tails (positive shape parameter)
        # Simulate from GPD with c=0.2 (heavy tail)
        extreme_returns = np.random.standard_t(3, 1000) * 0.02
        extreme_returns = extreme_returns[extreme_returns < -0.02]  # Take left tail

        if len(extreme_returns) > 30:
            par = gpd.fit(extreme_returns)
            assert isinstance(par, dict)
            assert 'c' in par
            assert 'scale' in par and par['scale'] > 0

    def test_fit_validation_empty_array(self):
        """Test that fit raises error for empty array."""
        with pytest.raises(ValueError, match="Returns cannot be empty"):
            gpd.fit([])

    def test_fit_validation_nan_values(self):
        """Test that fit raises error for NaN values."""
        returns = [0.01, np.nan, 0.02]
        with pytest.raises(ValueError, match="Returns must be finite"):
            gpd.fit(returns)

    def test_fit_validation_inf_values(self):
        """Test that fit raises error for Inf values."""
        returns = [0.01, np.inf, 0.02]
        with pytest.raises(ValueError, match="Returns must be finite"):
            gpd.fit(returns)


class TestGPDCDF:
    """Test GPD CDF functionality."""

    def setup_method(self):
        """Set up test parameters."""
        self.par = {'c': 0.1, 'loc': 0.0, 'scale': 1.0}

    def test_cdf_scalar_input(self):
        """Test CDF with scalar input."""
        result = gpd.cdf(0.5, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert 0 <= result[0] <= 1

    def test_cdf_array_input(self):
        """Test CDF with array input."""
        x = np.array([-1.0, 0.0, 1.0])
        result = gpd.cdf(x, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all((result >= 0) & (result <= 1))

    def test_cdf_monotonic(self):
        """Test that CDF is monotonically increasing."""
        x = np.linspace(-2.0, 2.0, 10)
        result = gpd.cdf(x, self.par)
        diffs = np.diff(result)
        assert np.all(diffs >= -1e-10), "CDF should be monotonically increasing"

    def test_cdf_boundaries(self):
        """Test CDF at extreme values."""
        # Very negative should be near 0
        result_low = gpd.cdf(-5.0, self.par)
        assert result_low[0] < 0.01, "CDF at very negative should be near 0"

        # Very positive should approach 1
        result_high = gpd.cdf(5.0, self.par)
        assert result_high[0] > 0.99, "CDF at very positive should approach 1"


class TestGPDPDF:
    """Test GPD PDF functionality."""

    def setup_method(self):
        """Set up test parameters."""
        self.par = {'c': 0.1, 'loc': 0.0, 'scale': 1.0}

    def test_pdf_scalar_input(self):
        """Test PDF with scalar input."""
        result = gpd.pdf(0.5, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert result[0] > 0
        assert np.isfinite(result[0])

    def test_pdf_array_input(self):
        """Test PDF with array input."""
        x = np.array([0.5, 1.0, 2.0])
        result = gpd.pdf(x, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all(result > 0)
        assert np.all(np.isfinite(result))

    def test_pdf_positive(self):
        """Test that PDF values are always positive."""
        returns = generate_extreme_returns(seed=42, size=100)
        par = gpd.fit(returns)

        x = np.linspace(0, 2, 10)
        result = gpd.pdf(x, par)

        assert np.all(result > 0), "All PDF values should be positive"
        assert np.all(np.isfinite(result)), "All PDF values should be finite"


class TestGPDPPF:
    """Test GPD PPF functionality."""

    def setup_method(self):
        """Set up test parameters."""
        self.par = {'c': 0.1, 'loc': 0.0, 'scale': 1.0}

    def test_ppf_scalar_input(self):
        """Test PPF with scalar input."""
        result = gpd.ppf(0.5, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert np.isfinite(result[0])

    def test_ppf_array_input(self):
        """Test PPF with array input."""
        q = np.array([0.1, 0.5, 0.9])
        result = gpd.ppf(q, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all(np.isfinite(result))

    @pytest.mark.parametrize("q", [0.05, 0.1, 0.5, 0.9, 0.95])
    def test_ppf_cdf_roundtrip(self, q):
        """Test that PPF and CDF are inverses."""
        ppf_result = gpd.ppf(q, self.par)
        cdf_result = gpd.cdf(ppf_result, self.par)
        assert np.isclose(cdf_result[0], q, rtol=1e-5), \
            f"PPF->CDF roundtrip failed: expected {q}, got {cdf_result[0]}"

    def test_ppf_boundaries(self):
        """Test PPF at boundary values."""
        # Very low quantile should give very negative result
        result_low = gpd.ppf(0.001, self.par)
        assert result_low[0] < 0

        # Very high quantile should give positive result
        result_high = gpd.ppf(0.999, self.par)
        assert result_high[0] > 0


class TestGPDRisk:
    """Test GPD risk calculations."""

    def setup_method(self):
        """Set up test data and parameters."""
        self.returns = generate_extreme_returns(seed=42, size=500)
        self.par = gpd.fit(self.returns)

    @pytest.mark.parametrize("alpha", [0.1, 0.05, 0.01, 0.001])
    def test_var_at_different_levels(self, alpha):
        """Test VaR calculation at different confidence levels."""
        var = gpd.var(alpha, self.par)

        assert isinstance(var, float), f"VaR should return float at alpha={alpha}"
        assert np.isfinite(var), f"VaR should be finite, got {var}"
        assert not np.isnan(var), f"VaR should not be NaN"

        # Lower alpha should produce more extreme VaR
        if alpha < 0.05:
            var_5 = gpd.var(0.05, self.par)
            assert var <= var_5, "Lower alpha should produce more extreme VaR"

    @pytest.mark.parametrize("alpha", [0.1, 0.05, 0.01, 0.001])
    def test_es_at_different_levels(self, alpha):
        """Test ES calculation at different confidence levels."""
        es = gpd.es(alpha, self.par)

        assert isinstance(es, float), f"ES should return float at alpha={alpha}"
        assert np.isfinite(es), f"ES should be finite, got {es}"
        assert not np.isnan(es), f"ES should not be NaN"

    def test_var_es_consistency(self):
        """Test that VaR and ES follow risk measure properties."""
        for alpha in [0.05, 0.01]:
            var = gpd.var(alpha, self.par)
            es = gpd.es(alpha, self.par)

            # For loss distributions (negative returns), ES should be more extreme than VaR
            assert es <= var, f"ES ({es}) should be <= VaR ({var}) for alpha={alpha}"

    def test_var_alpha_validation(self):
        """Test that var validates alpha parameter."""
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            gpd.var(1.5, self.par)

        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            gpd.var(-0.1, self.par)

    def test_es_alpha_validation(self):
        """Test that es validates alpha parameter."""
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            gpd.es(1.5, self.par)

        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            gpd.es(-0.1, self.par)

    def test_var_handles_scalar_return(self):
        """Test that var handles scalar return from ppf (regression test)."""
        var = gpd.var(0.05, self.par)
        assert isinstance(var, float)
        assert not np.isnan(var)


class TestGPDSpecialCases:
    """Test GPD special cases and edge conditions."""

    def test_exponential_case(self):
        """Test GPD with c=0 (exponential distribution)."""
        # When c=0, GPD reduces to exponential distribution
        par = {'c': 0.0, 'loc': 0.0, 'scale': 1.0}

        # Should work without errors
        pdf_val = gpd.pdf(0.5, par)
        cdf_val = gpd.cdf(0.5, par)
        ppf_val = gpd.ppf(0.5, par)

        assert isinstance(pdf_val, np.ndarray)
        assert isinstance(cdf_val, np.ndarray)
        assert isinstance(ppf_val, np.ndarray)
        assert pdf_val[0] > 0
        assert 0 <= cdf_val[0] <= 1

    def test_heavy_tail_case(self):
        """Test GPD with c > 0 (heavy tail)."""
        # Positive shape parameter gives heavy tail
        par = {'c': 0.3, 'loc': 0.0, 'scale': 1.0}

        # Should be able to calculate VaR and ES
        var = gpd.var(0.05, par)
        es = gpd.es(0.05, par)

        assert isinstance(var, float)
        assert isinstance(es, float)
        assert np.isfinite(var)
        assert np.isfinite(es)

    def test_bounded_case(self):
        """Test GPD with c < 0 (bounded tail)."""
        # Negative shape parameter gives bounded tail
        par = {'c': -0.3, 'loc': 0.0, 'scale': 1.0}

        # Should be able to calculate VaR and ES
        var = gpd.var(0.05, par)
        es = gpd.es(0.05, par)

        assert isinstance(var, float)
        assert isinstance(es, float)
        assert np.isfinite(var)
        assert np.isfinite(es)



"""Student's T distribution - pure functions."""

import numpy as np
from scipy.stats import t as student_t
from scipy.integrate import quad
from .util import validate_returns, validate_alpha


def fit(returns):
    """Fit Student's T distribution using MLE.

    Parameters
    ----------
    returns : array-like
        Array of returns

    Returns
    -------
    dict
        Fitted parameters: {'df': float, 'loc': float, 'scale': float}
    """
    returns = validate_returns(returns)
    df, loc, scale = student_t.fit(returns)
    return {'df': df, 'loc': loc, 'scale': scale}


def pdf(x, par):
    """Probability density function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate PDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PDF values
    """
    return student_t.pdf(x, **par)


def cdf(x, par):
    """Cumulative distribution function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate CDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        CDF values
    """
    return student_t.cdf(x, **par)


def ppf(q, par):
    """Percent point function (inverse CDF).

    Parameters
    ----------
    q : array-like
        Quantiles to evaluate
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PPF values
    """
    return student_t.ppf(q, **par)


def var(alpha, par):
    """Value at Risk at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        VaR value
    """
    validate_alpha(alpha)
    result = ppf(alpha, par)
    return float(result) if np.isscalar(result) else float(result[0])


def es(alpha, par):
    """Expected Shortfall at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        ES value
    """
    validate_alpha(alpha)

    # For Student's T, we can use numerical integration with better bounds
    var_val = var(alpha, par)

    # Use finite integration bounds to avoid quad issues
    lower_bound = var_val - 10 * par['scale']
    if np.isinf(lower_bound):
        lower_bound = -10

    def integrand(x):
        pdf_val = pdf(x, par)
        if np.isscalar(x):
            return x * float(pdf_val)
        return x * pdf_val

    es_val, _ = quad(integrand, lower_bound, var_val, limit=100)
    return es_val / alpha



"""Common utilities for testing pyreto distributions."""

import numpy as np


def generate_test_returns(seed=42, size=200):
    """Generate test returns data with realistic properties."""
    np.random.seed(seed)
    # Mix of normal and some extreme values to simulate fat tails
    normal_returns = np.random.normal(0, 0.02, size)
    extreme_prob = 0.02  # 2% chance of extreme returns
    extreme_mask = np.random.random(size) < extreme_prob
    extreme_values = np.random.choice([-0.05, 0.05], size=size) * np.random.random(size)
    
    returns = normal_returns.copy()
    returns[extreme_mask] = extreme_values[extreme_mask]
    
    return returns


def check_var_es_properties(var, es, alpha):
    """Check that VaR and ES follow basic risk measure properties."""
    # ES should be more extreme (worse) than VaR for same alpha
    if alpha < 0.5:
        assert es <= var, f"ES ({es}) should be <= VaR ({var}) for alpha={alpha}"
    else:
        assert es >= var, f"ES ({es}) should be >= VaR ({var}) for alpha={alpha}"
    
    # VaR should be a float
    assert isinstance(var, float), f"VaR should be float, got {type(var)}"
    assert isinstance(es, float), f"ES should be float, got {type(es)}"


def check_distribution_properties(distro_module, returns, tol=0.1):
    """Check basic properties of fitted distributions."""
    par = distro_module.fit(returns)
    
    # Check that parameters were fit and are valid numbers
    assert isinstance(par, dict), "fit() should return dict"
    assert len(par) > 0, "Parameters should not be empty"
    
    for key, val in par.items():
        assert isinstance(val, (float, np.floating)), \
            f"Parameter {key} should be float-like, got {type(val)}"
        assert np.isfinite(val), f"Parameter {key} should be finite, got {val}"
    
    return par



"""Pyreto: Tail risk management library with Monte Carlo simulation (classless functional API)."""

__version__ = "0.5.0"

# Make distributions available at pyreto.*
from .distributions import student_t
from .distributions import alpha_stable
from .distributions import nig
from .distributions import gpd

__all__ = ["student_t", "alpha_stable", "nig", "gpd"]



"""Minimal Monte Carlo engine (CPU-only, single-threaded)."""

import numpy as np
import polars as pl
from scipy import stats

from . import student_t, nig, alpha_stable


_MARGINAL_INVERTERS = {
    "student_t": student_t.ppf,
    "nig": nig.ppf,
    "alpha_stable": alpha_stable.ppf,
}


def simulate(copula, margins, n_draws, seed=None):
    """
    Simulate correlated log-returns via copula + margins.

    Parameters
    ----------
    copula : str
        "gaussian" or "t"
    margins : dict
        {"asset": {"type": "student_t", "par": {...}}, ...}
    n_draws : int
        Number of draws (>0)
    seed : int, optional
        Random seed

    Returns
    -------
    polars.DataFrame
        Shape (n_draws, n_assets) with simulated log-returns
    """
    assets = list(margins)
    if not assets:
        raise ValueError("margins dict cannot be empty")
    if n_draws <= 0:
        raise ValueError("n_draws must be positive")

    rng = np.random.default_rng(seed)
    n_assets = len(assets)

    # Uniform draws from copula (identity correlation)
    if copula == "gaussian":
        normals = rng.standard_normal((n_draws, n_assets))
        uniforms = stats.norm.cdf(normals)
    elif copula == "t":
        dof = 5  # canonical choice, matches typical risk-model dofs
        mvt = stats.multivariate_t(loc=np.zeros(n_assets),
                                   shape=np.eye(n_assets),
                                   df=dof)
        t_draws = mvt.rvs(size=n_draws, random_state=rng)
        uniforms = stats.t.cdf(t_draws, df=dof)
    else:
        raise ValueError("copula must be 'gaussian' or 't'")

    # Apply marginal inverse CDFs
    out = {}
    for i, asset in enumerate(assets):
        margin = margins[asset]
        inv = _MARGINAL_INVERTERS.get(margin["type"])
        if inv is None:
            raise ValueError(f"unknown margin type: {margin['type']}")
        out[asset] = inv(uniforms[:, i], margin["par"])

    return pl.DataFrame(out)


def simulate_portfolio(weights, mc_returns):
    """
    Portfolio log-returns from Monte Carlo simulation.

    Parameters
    ----------
    weights : numpy.ndarray
        Portfolio weights, must sum to 1 and be non-negative
    mc_returns : polars.DataFrame
        Simulated returns from simulate()

    Returns
    -------
    polars.Series
        Portfolio log-return series
    """
    w = np.asarray(weights, dtype=float)

    if w.ndim != 1:
        raise ValueError("weights must be 1-dimensional")
    if np.any(w < 0):
        raise ValueError("weights must be non-negative")
    if not np.allclose(w.sum(), 1.0):
        raise ValueError("weights must sum to 1")
    if len(w) != len(mc_returns.columns):
        raise ValueError("weights length must match number of assets")

    return pl.Series("portfolio_return", mc_returns.to_numpy() @ w)



"""Generalised Pareto Distribution - pure functions for Peaks Over Threshold."""

import numpy as np
from scipy.stats import genpareto
from scipy.integrate import quad

from .util import validate_returns, validate_alpha


def fit(returns):
    """Fit Generalised Pareto Distribution using MLE.

    Parameters
    ----------
    returns : array-like
        Array of returns (extreme values / exceedances)

    Returns
    -------
    dict
        Fitted parameters: {'c': float, 'loc': float, 'scale': float}
        c: shape parameter
        loc: location parameter (threshold)
        scale: scale parameter
    """
    returns = validate_returns(returns)

    # For GPD, we typically model exceedances over a threshold
    # Here we let scipy handle the fitting
    c, loc, scale = genpareto.fit(returns)

    return {'c': c, 'loc': loc, 'scale': scale}


def pdf(x, par):
    """Probability density function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate PDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PDF values
    """
    result = genpareto.pdf(x, **par)
    if np.ndim(result) == 0:
        return np.array([result])
    return result


def cdf(x, par):
    """Cumulative distribution function.

    Parameters
    ----------
    ----------
    x : array-like
        Points at which to evaluate CDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        CDF values
    """
    result = genpareto.cdf(x, **par)
    if np.ndim(result) == 0:
        return np.array([result])
    return result


def ppf(q, par):
    """Percent point function (inverse CDF).

    Parameters
    ----------
    q : array-like
        Quantiles to evaluate
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PPF values
    """
    result = genpareto.ppf(q, **par)
    if np.ndim(result) == 0:
        return np.array([result])
    return result


def var(alpha, par):
    """Value at Risk at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        VaR value
    """
    validate_alpha(alpha)
    result = ppf(alpha, par)
    if np.ndim(result) == 0:
        return float(result)
    else:
        return float(result[0])


def es(alpha, par):
    """Expected Shortfall at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        ES value
    """
    validate_alpha(alpha)
    var_val = var(alpha, par)

    def integrand(x):
        return x * pdf(x, par)[0]

    # Use finite integration bounds for GPD
    # For c > 0, distribution has finite upper bound at loc - scale/c
    # For c <= 0, distribution is unbounded to the right
    if par['c'] > 0:
        upper_bound = par['loc'] - par['scale'] / par['c']
        es_val, _ = quad(integrand, -np.inf, min(var_val, upper_bound - 1e-6))
    else:
        es_val, _ = quad(integrand, -np.inf, var_val)

    return es_val / alpha



"""Alpha-Stable distribution - pure functions."""

import numpy as np
from scipy.optimize import minimize
from scipy.integrate import quad
from scipy.optimize import brentq
from scipy.stats import norm
from .util import validate_returns, validate_alpha


def _char_func(t, alpha, beta, c, mu):
    """Characteristic function."""
    if alpha == 1:
        return np.exp(1j * mu * t - c * np.abs(t) *
                      (1 + 1j * beta * np.sign(t) * np.log(np.abs(t))))
    else:
        return np.exp(1j * mu * t - c**alpha * np.abs(t)**alpha *
                      (1 - 1j * beta * np.sign(t) * np.tan(np.pi * alpha / 2)))


def _pdf_single(x, alpha, beta, c, mu):
    """PDF for single point."""
    if alpha == 2.0:
        return norm.pdf(x, loc=mu, scale=c)

    t = np.linspace(-50, 50, 1024)
    cf = _char_func(t, alpha, beta, c, mu - x)
    dt = t[1] - t[0]
    cf_sym = np.concatenate([cf, np.conj(cf[-1:0:-1])])
    pdf_component = np.fft.ifft(cf_sym).real[:len(t)] * len(t) / (2 * np.pi)
    return max(abs(pdf_component[0]), 1e-10)


def fit(returns):
    """Fit Alpha-Stable distribution using MLE.

    Parameters
    ----------
    returns : array-like
        Array of returns

    Returns
    -------
    dict
        Fitted parameters: {'alpha': float, 'beta': float, 'c': float, 'mu': float}
    """
    returns = validate_returns(returns)

    def neg_loglik(params):
        alpha, beta, c, mu = params
        if not (0.1 < alpha <= 2.0) or not (-1 <= beta <= 1) or c <= 0:
            return 1e10
        try:
            pdf_vals = [_pdf_single(x, alpha, beta, c, mu) for x in returns]
            if any(p <= 0 for p in pdf_vals):
                return 1e10
            return -np.sum(np.log(pdf_vals))
        except:
            return 1e10

    mu0 = np.mean(returns)
    c0 = np.std(returns) / np.sqrt(2)
    alpha0 = 1.8
    beta0 = 0.0
    x0 = np.array([alpha0, beta0, c0, mu0])
    bounds = [(0.1, 2.0), (-0.999, 0.999), (1e-6, 10.0), (None, None)]

    res = minimize(neg_loglik, x0, method='Nelder-Mead',
                   options={'maxiter': 500, 'xatol': 1e-6, 'fatol': 1e-6})

    alpha, beta, c, mu = res.x
    return {'alpha': alpha, 'beta': beta, 'c': c, 'mu': mu}


def pdf(x, par):
    """Probability density function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate PDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PDF values
    """
    x = np.atleast_1d(x)
    results = np.zeros_like(x)
    for i, xi in enumerate(x):
        results[i] = _pdf_single(
            xi, par['alpha'], par['beta'], par['c'], par['mu'])
    return results


def cdf(x, par):
    """Cumulative distribution function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate CDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        CDF values
    """
    x = np.atleast_1d(x)
    results = np.zeros_like(x)
    for i, xi in enumerate(x):
        results[i], _ = quad(lambda xx: pdf(xx, par), -np.inf, xi)
    return results


def ppf(q, par):
    """Percent point function (inverse CDF).

    Parameters
    ----------
    q : array-like
        Quantiles to evaluate
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PPF values
    """
    q = np.atleast_1d(q)
    results = np.zeros_like(q)

    for i, qi in enumerate(q):
        def f(x):
            return cdf(np.array([x]), par)[0] - qi
        results[i] = brentq(f, -10, 10)

    return results


def var(alpha, par):
    """Value at Risk at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        VaR value
    """
    validate_alpha(alpha)
    return ppf(alpha, par)[0]


def es(alpha, par):
    """Expected Shortfall at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        ES value
    """
    validate_alpha(alpha)
    var_val = var(alpha, par)

    def integrand(x):
        return x * pdf(x, par)[0]

    es_val, _ = quad(integrand, -np.inf, var_val)
    return es_val / alpha



"""Unit tests for Student's T distribution."""

import pytest
import numpy as np
import sys
import os

# Add src to path for imports
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))

from pyreto import student_t
from test_common import generate_test_returns, check_var_es_properties, check_distribution_properties


class TestStudentTFit:
    """Test Student's T fitting functionality."""
    
    def test_fit_basic(self):
        """Test basic fitting functionality."""
        returns = generate_test_returns(seed=1, size=500)
        par = check_distribution_properties(student_t, returns)
        
        # Check specific parameters for Student's T
        assert 'df' in par, "Student's T should have 'df' parameter"
        assert 'loc' in par, "Student's T should have 'loc' parameter"
        assert 'scale' in par, "Student's T should have 'scale' parameter"
        
        assert par['df'] > 0, f"Degrees of freedom should be positive, got {par['df']}"
        assert par['scale'] > 0, f"Scale should be positive, got {par['scale']}"
    
    def test_fit_with_leptokurtic_data(self):
        """Test fitting with fat-tailed data."""
        np.random.seed(123)
        # Generate data with heavier tails
        normal = np.random.normal(0, 0.01, 1000)
        extreme = np.random.choice([-0.1, 0.1], size=50)
        returns = np.concatenate([normal, extreme])
        
        par = student_t.fit(returns)
        # Fat-tailed data should result in lower df (degrees of freedom)
        assert 0 < par['df'] < 10, f"Expected low df for fat tails, got {par['df']}"
    
    def test_fit_validation_empty_array(self):
        """Test that fit raises error for empty array."""
        with pytest.raises(ValueError, match="Returns cannot be empty"):
            student_t.fit([])
    
    def test_fit_validation_nan_values(self):
        """Test that fit raises error for NaN values."""
        returns = [0.01, np.nan, 0.02]
        with pytest.raises(ValueError, match="Returns must be finite"):
            student_t.fit(returns)
    
    def test_fit_validation_inf_values(self):
        """Test that fit raises error for Inf values."""
        returns = [0.01, np.inf, 0.02]
        with pytest.raises(ValueError, match="Returns must be finite"):
            student_t.fit(returns)


class TestStudentTRisk:
    """Test Student's T risk calculations."""
    
    def setup_method(self):
        """Set up test data and parameters."""
        self.returns = generate_test_returns(seed=42, size=500)
        self.par = student_t.fit(self.returns)
    
    @pytest.mark.parametrize("alpha", [0.1, 0.05, 0.01, 0.001])
    def test_var_at_different_levels(self, alpha):
        """Test VaR calculation at different confidence levels."""
        var = student_t.var(alpha, self.par)
        
        assert isinstance(var, float), f"VaR should return float, got {type(var)}"
        assert np.isfinite(var), f"VaR should be finite, got {var}"
        
        # Lower alpha should produce more extreme VaR
        if alpha < 0.05:
            var_5 = student_t.var(0.05, self.par)
            assert var <= var_5, "Lower alpha should produce more extreme VaR"
    
    @pytest.mark.parametrize("alpha", [0.1, 0.05, 0.01, 0.001])
    def test_es_at_different_levels(self, alpha):
        """Test ES calculation at different confidence levels."""
        es = student_t.es(alpha, self.par)
        
        assert isinstance(es, float), f"ES should return float, got {type(es)}"
        assert np.isfinite(es), f"ES should be finite, got {es}"
    
    def test_var_es_consistency(self):
        """Test that VaR and ES follow risk measure properties."""
        for alpha in [0.05, 0.01]:
            var = student_t.var(alpha, self.par)
            es = student_t.es(alpha, self.par)
            check_var_es_properties(var, es, alpha)
    
    def test_var_alpha_validation(self):
        """Test that var validates alpha parameter."""
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            student_t.var(1.5, self.par)
        
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            student_t.var(-0.1, self.par)
    
    def test_es_alpha_validation(self):
        """Test that es validates alpha parameter."""
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            student_t.es(1.5, self.par)
        
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            student_t.es(-0.1, self.par)
    
    def test_var_handles_scalar_return(self):
        """Test that var handles scalar return from ppf (regression test)."""
        # This was the bug: ppf returning scalar caused IndexError
        var = student_t.var(0.05, self.par)
        assert isinstance(var, float)
        assert not np.isnan(var)
    
    def test_var_handles_0d_array(self):
        """Test that var handles 0-dimensional array return from ppf."""
        # Some internal functions return 0-d arrays
        alpha = np.float64(0.05)  # This can cause 0-d array from scipy
        var = student_t.var(alpha, self.par)
        assert isinstance(var, float)
        assert not np.isnan(var)


class TestStudentTPDF:
    """Test Student's T PDF functionality."""
    
    def setup_method(self):
        """Set up test parameters."""
        self.par = {'df': 5.0, 'loc': 0.0, 'scale': 1.0}
    
    def test_pdf_scalar_input(self):
        """Test PDF with scalar input."""
        result = student_t.pdf(0.0, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert result[0] > 0
        assert np.isfinite(result[0])
    
    def test_pdf_array_input(self):
        """Test PDF with array input."""
        x = np.array([-1.0, 0.0, 1.0])
        result = student_t.pdf(x, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all(result > 0)
        assert np.all(np.isfinite(result))
    
    def test_pdf_symmetry(self):
        """Test that PDF is symmetric for location=0."""
        x = np.array([-1.0, 1.0])
        result = student_t.pdf(x, self.par)
        assert np.isclose(result[0], result[1], rtol=1e-5)
    
    def test_pdf_location_shift(self):
        """Test that location parameter shifts the distribution."""
        par_shifted = {'df': 5.0, 'loc': 2.0, 'scale': 1.0}
        result_center = student_t.pdf(2.0, self.par)
        result_shifted = student_t.pdf(4.0, par_shifted)
        # Value at 2.0 for centered should equal value at 4.0 for shifted by 2
        assert np.isclose(result_center[0], result_shifted[0], rtol=1e-5)


class TestStudentTCDF:
    """Test Student's T CDF functionality."""
    
    def setup_method(self):
        """Set up test parameters."""
        self.par = {'df': 5.0, 'loc': 0.0, 'scale': 1.0}
    
    def test_cdf_scalar_input(self):
        """Test CDF with scalar input."""
        result = student_t.cdf(0.0, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert 0 <= result[0] <= 1
    
    def test_cdf_array_input(self):
        """Test CDF with array input."""
        x = np.array([-1.0, 0.0, 1.0])
        result = student_t.cdf(x, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all((result >= 0) & (result <= 1))
    
    def test_cdf_monotonic(self):
        """Test that CDF is monotonically increasing."""
        x = np.linspace(-3.0, 3.0, 7)
        result = student_t.cdf(x, self.par)
        assert np.all(np.diff(result) >= 0), "CDF should be monotonically increasing"


class TestStudentTPPF:
    """Test Student's T PPF functionality."""
    
    def setup_method(self):
        """Set up test parameters."""
        self.par = {'df': 5.0, 'loc': 0.0, 'scale': 1.0}
    
    def test_ppf_scalar_input(self):
        """Test PPF with scalar input."""
        result = student_t.ppf(0.5, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert np.isfinite(result[0])
    
    def test_ppf_array_input(self):
        """Test PPF with array input."""
        q = np.array([0.1, 0.5, 0.9])
        result = student_t.ppf(q, self.par)
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all(np.isfinite(result))
    
    @pytest.mark.parametrize("q", [0.01, 0.05, 0.1, 0.5, 0.9, 0.95, 0.99])
    def test_ppf_cdf_roundtrip(self, q):
        """Test that PPF and CDF are inverses."""
        ppf_result = student_t.ppf(q, self.par)
        cdf_result = student_t.cdf(ppf_result, self.par)
        assert np.isclose(cdf_result[0], q, rtol=1e-5), \
            f"PPF->CDF roundtrip failed: expected {q}, got {cdf_result[0]}"
    
    def test_ppf_boundaries(self):
        """Test PPF at boundary values."""
        # Extremely small quantile should give very negative result
        result_low = student_t.ppf(0.001, self.par)
        assert result_low[0] < -2.0
        
        # Extremely high quantile should give very positive result
        result_high = student_t.ppf(0.999, self.par)
        assert result_high[0] > 2.0



"""Normal-Inverse Gaussian distribution - pure functions."""

import numpy as np
from scipy.optimize import minimize
from scipy.special import kv as bessel_k
from scipy.integrate import quad
from scipy.optimize import brentq
from .util import validate_returns, validate_alpha


def _pdf_single(x, alpha, beta, delta, mu):
    """PDF for single point."""
    gamma = np.sqrt(alpha**2 - beta**2)
    z = (x - mu) / delta
    factor = alpha * delta * np.sqrt(1 + z**2)
    numerator = alpha * delta * bessel_k(1, factor)
    denominator = np.pi * np.sqrt(1 + z**2)
    return (numerator / denominator) * np.exp(delta * gamma + beta * (x - mu))


def fit(returns):
    """Fit Normal-Inverse Gaussian distribution using MLE.

    Parameters
    ----------
    returns : array-like
        Array of returns

    Returns
    -------
    dict
        Fitted parameters: {'alpha': float, 'beta': float, 
                           'delta': float, 'mu': float, 'gamma': float}
    """
    returns = validate_returns(returns)

    def neg_loglik(params):
        alpha, beta, delta, mu = params
        if alpha <= 0 or delta <= 0 or abs(beta) >= alpha:
            return 1e10
        try:
            pdf_vals = [_pdf_single(x, alpha, beta, delta, mu)
                        for x in returns]
            return -np.sum(np.log(pdf_vals))
        except:
            return 1e10

    mu0 = np.mean(returns)
    delta0 = np.std(returns) * 0.8
    alpha0 = 2.0
    beta0 = 0.0
    x0 = np.array([alpha0, beta0, delta0, mu0])
    bounds = [(1e-4, 50), (-49, 49), (1e-4, 50), (None, None)]

    res = minimize(neg_loglik, x0, method='SLSQP', bounds=bounds,
                   options={'ftol': 1e-8, 'maxiter': 1000})

    alpha, beta, delta, mu = res.x
    return {'alpha': alpha, 'beta': beta, 'delta': delta, 'mu': mu,
            'gamma': np.sqrt(alpha**2 - beta**2)}


def pdf(x, par):
    """Probability density function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate PDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PDF values
    """
    x = np.atleast_1d(x)
    results = np.zeros_like(x)
    for i, xi in enumerate(x):
        results[i] = _pdf_single(
            xi, par['alpha'], par['beta'], par['delta'], par['mu'])
    return results


def cdf(x, par):
    """Cumulative distribution function.

    Parameters
    ----------
    x : array-like
        Points at which to evaluate CDF
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        CDF values
    """
    x = np.atleast_1d(x)
    results = np.zeros_like(x)
    for i, xi in enumerate(x):
        results[i], _ = quad(lambda xx: pdf(xx, par), -np.inf, xi)
    return results


def ppf(q, par):
    """Percent point function (inverse CDF).

    Parameters
    ----------
    q : array-like
        Quantiles to evaluate
    par : dict
        Parameters from fit()

    Returns
    -------
    ndarray
        PPF values
    """
    q = np.atleast_1d(q)
    results = np.zeros_like(q)

    for i, qi in enumerate(q):
        def f(x):
            return cdf(np.array([x]), par)[0] - qi
        results[i] = brentq(f, -10, 10)

    return results


def var(alpha, par):
    """Value at Risk at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        VaR value
    """
    validate_alpha(alpha)
    return ppf(alpha, par)[0]


def es(alpha, par):
    """Expected Shortfall at significance level alpha.

    Parameters
    ----------
    alpha : float
        Significance level (0 < alpha < 1)
    par : dict
        Parameters from fit()

    Returns
    -------
    float
        ES value
    """
    validate_alpha(alpha)
    var_val = var(alpha, par)

    def integrand(x):
        return x * pdf(x, par)[0]

    es_val, _ = quad(integrand, -np.inf, var_val)
    return es_val / alpha



"""Tests for minimalist Monte Carlo engine."""

import sys
import os

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

import numpy as np
import polars as pl

from pyreto import mc, student_t, nig


def test_gaussian_copula_three_assets():
    """3-asset Gaussian copula, mean portfolio return ≈ 0."""
    margins = {
        "SPY": {"type": "student_t", "par": {"df": 4.0, "loc": 0.0, "scale": 0.01}},
        "TLT": {"type": "student_t", "par": {"df": 5.0, "loc": 0.0, "scale": 0.008}},
        "GLD": {"type": "student_t", "par": {"df": 6.0, "loc": 0.0, "scale": 0.012}},
    }

    returns = mc.simulate("gaussian", margins, n_draws=20_000, seed=42)
    assert isinstance(returns, pl.DataFrame)
    assert returns.shape == (20_000, 3)
    assert list(returns.columns) == ["SPY", "TLT", "GLD"]

    # Equally weighted portfolio
    weights = np.array([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])
    port = mc.simulate_portfolio(weights, returns)
    mean_ret = port.mean()

    assert abs(mean_ret) < 1e-2, f"Mean portfolio return {mean_ret} exceeds tolerance"


def test_t_copula_heavier_tails():
    """t-copula with dof=5 produces heavier tails than Gaussian."""
    margins = {
        "SPY": {"type": "student_t", "par": {"df": 4.0, "loc": 0.0, "scale": 0.01}}
    }

    # Gaussian
    gauss = mc.simulate("gaussian", margins, n_draws=50_000, seed=123)
    gauss_kurt = gauss["SPY"].kurtosis()

    # t-copula
    t_draws = mc.simulate("t", margins, n_draws=50_000, seed=123)
    t_kurt = t_draws["SPY"].kurtosis()

    assert (
        t_kurt > gauss_kurt
    ), "t-copula should produce heavier tails than Gaussian"
    assert t_kurt > 3.0, "t-copula tail kurtosis should exceed 3"


def test_reproducibility():
    """Same seed → identical draws."""
    margins = {
        "SPY": {"type": "student_t", "par": {"df": 4.0, "loc": 0.0, "scale": 0.01}}
    }

    r1 = mc.simulate("gaussian", margins, n_draws=10_000, seed=42)
    r2 = mc.simulate("gaussian", margins, n_draws=10_000, seed=42)

    assert (r1["SPY"] == r2["SPY"]).all(), "Same seed must give identical draws"


def test_simulate_portfolio_shape():
    """Portfolio shape and weight validation."""
    margins = {
        "A": {"type": "student_t", "par": {"df": 5.0, "loc": 0.0, "scale": 0.02}},
        "B": {"type": "student_t", "par": {"df": 5.0, "loc": 0.0, "scale": 0.02}},
    }

    returns = mc.simulate("gaussian", margins, n_draws=1_000, seed=1)
    weights = np.array([0.6, 0.4])

    port = mc.simulate_portfolio(weights, returns)
    assert isinstance(port, pl.Series)
    assert len(port) == 1_000
    assert port.name == "portfolio_return"


def test_nig_margins():
    """NIG margins work in simulation."""
    # Fit NIG to some synthetic data
    np.random.seed(7)
    syn = np.random.normal(0, 0.02, 5000)
    par = nig.fit(syn)

    margins = {"asset1": {"type": "nig", "par": par}}
    draws = mc.simulate("gaussian", margins, n_draws=5_000, seed=77)

    assert draws.shape == (5_000, 1)
    assert not draws["asset1"].is_null().any()
    assert draws["asset1"].std() > 0


def test_negative_weights_raise():
    """Negative weights are rejected."""
    margins = {"A": {"type": "student_t", "par": {"df": 5.0, "loc": 0, "scale": 0.02}}}
    returns = mc.simulate("gaussian", margins, n_draws=100, seed=1)

    try:
        mc.simulate_portfolio(np.array([-0.5, 1.5]), returns)
        assert False, "Should have raised ValueError"
    except ValueError as e:
        assert "non-negative" in str(e)


def test_weights_must_sum_to_one():
    """Weights not summing to 1 are rejected."""
    margins = {"A": {"type": "student_t", "par": {"df": 5.0, "loc": 0, "scale": 0.02}}}
    returns = mc.simulate("gaussian", margins, n_draws=100, seed=1)

    try:
        mc.simulate_portfolio(np.array([0.5, 0.5]), returns)
        assert False, "Should have raised ValueError"
    except ValueError as e:
        assert "sum to 1" in str(e)


def test_single_asset_portfolio():
    """Single asset portfolio returns identical to asset."""
    margins = {
        "ONLY": {"type": "student_t", "par": {"df": 5.0, "loc": 0, "scale": 0.02}}
    }

    returns = mc.simulate("gaussian", margins, n_draws=1_000, seed=99)
    port = mc.simulate_portfolio(np.array([1.0]), returns)

    assert (port == returns["ONLY"]).all()


def test_t_copula_dof5_empirical_tail():
    """Empirical check: t-copula gives more 5 % VaR exceedances."""
    margins = {
        "X": {"type": "student_t", "par": {"df": 4.0, "loc": 0.0, "scale": 0.01}}
    }

    gauss = mc.simulate("gaussian", margins, n_draws=100_000, seed=10)
    t_draws = mc.simulate("t", margins, n_draws=100_000, seed=10)

    var_gauss = np.percentile(gauss["X"], 5)
    var_t = np.percentile(t_draws["X"], 5)

    # t-copula should produce more extreme VaR
    assert var_t < var_gauss, "t-copula VaR should be more extreme than Gaussian"


if __name__ == "__main__":
    # Run all tests
    test_gaussian_copula_three_assets()
    print("✓ test_gaussian_copula_three_assets")

    test_t_copula_heavier_tails()
    print("✓ test_t_copula_heavier_tails")

    test_reproducibility()
    print("✓ test_reproducibility")

    test_simulate_portfolio_shape()
    print("✓ test_simulate_portfolio_shape")

    test_nig_margins()
    print("✓ test_nig_margins")

    test_negative_weights_raise()
    print("✓ test_negative_weights_raise")

    test_weights_must_sum_to_one()
    print("✓ test_weights_must_sum_to_one")

    test_single_asset_portfolio()
    print("✓ test_single_asset_portfolio")

    test_t_copula_dof5_empirical_tail()
    print("✓ test_t_copula_dof5_empirical_tail")

    print("\nAll MC tests passed!")



import numpy as np
import warnings
from typing import Sequence


def validate_returns(returns: Sequence[float] | np.ndarray) -> np.ndarray:
    """
    Convert `returns` to a 1-D float64 NumPy array and run basic checks.

    Returns
    -------
    np.ndarray
        Validated, ravelled copy/view of the input.

    Raises
    ------
    ValueError
        If the array is empty or contains non-finite values.
    """
    # NOTE: should this also warn if len(returns) below certain threshold? below certain length no statistical reasoning makes sense?
    if not isinstance(returns, np.ndarray):
        returns = np.asarray(returns, dtype=np.float64)
    else:
        returns = returns.astype(np.float64, copy=False)

    returns = returns.ravel()

    if returns.size == 0:
        raise ValueError("Returns cannot be empty")
    if not np.isfinite(returns).all():
        raise ValueError("Returns must be finite (no NaN or Inf)")
    if returns.size < 30:
        warnings.warn(
            "Statistical power low with fewer than 30 observations",
            UserWarning,
            stacklevel=2
        )

    return returns


def validate_alpha(alpha: float):
    if not (0 < alpha < 1):
        raise ValueError("Alpha must be between 0 and 1")



"""Unit tests for Alpha-Stable distribution."""

import pytest
import numpy as np
import sys
import os

# Add src to path for imports
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'src'))

from pyreto import alpha_stable
from test_common import generate_test_returns, check_var_es_properties, check_distribution_properties


class TestAlphaStableFit:
    """Test Alpha-Stable fitting functionality."""
    
    def test_fit_basic(self):
        """Test basic fitting functionality."""
        returns = generate_test_returns(seed=1, size=500)
        par = check_distribution_properties(alpha_stable, returns)
        
        # Check specific parameters for Alpha-Stable
        required_params = ['alpha', 'beta', 'c', 'mu']
        for param in required_params:
            assert param in par, f"Alpha-Stable should have '{param}' parameter"
        
        assert 0.1 < par['alpha'] <= 2.0, f"Alpha should be in (0.1, 2.0], got {par['alpha']}"
        assert -1 <= par['beta'] <= 1, f"Beta should be in [-1, 1], got {par['beta']}"
        assert par['c'] > 0, f"c (scale) should be positive, got {par['c']}"
    
    def test_fit_symmetric_case(self):
        """Test fitting symmetric data (beta should be near 0)."""
        np.random.seed(123)
        # Symmetric normal data
        returns = np.random.normal(0, 0.02, 1000)
        
        par = alpha_stable.fit(returns)
        # For symmetric data, beta should be close to 0
        assert abs(par['beta']) < 0.5, f"Beta should be near 0 for symmetric data, got {par['beta']}"
        # Alpha should be near 2 for normal-like data
        assert par['alpha'] > 1.5, f"Alpha should be high for normal-like data, got {par['alpha']}"
    
    def test_fit_validation_empty_array(self):
        """Test that fit raises error for empty array."""
        with pytest.raises(ValueError, match="Returns cannot be empty"):
            alpha_stable.fit([])
    
    def test_fit_validation_nan_values(self):
        """Test that fit raises error for NaN values."""
        returns = [0.01, np.nan, 0.02]
        with pytest.raises(ValueError, match="Returns must be finite"):
            alpha_stable.fit(returns)
    
    def test_fit_validation_inf_values(self):
        """Test that fit raises error for Inf values."""
        returns = [0.01, np.inf, 0.02]
        with pytest.raises(ValueError, match="Returns must be finite"):
            alpha_stable.fit(returns)


class TestAlphaStableRisk:
    """Test Alpha-Stable risk calculations."""
    
    def setup_method(self):
        """Set up test data and parameters."""
        self.returns = generate_test_returns(seed=42, size=500)
        self.par = alpha_stable.fit(self.returns)
    
    @pytest.mark.parametrize("alpha", [0.1, 0.05, 0.01, 0.001])
    def test_var_at_different_levels(self, alpha):
        """Test VaR calculation at different confidence levels."""
        var = alpha_stable.var(alpha, self.par)
        
        assert isinstance(var, float), f"VaR should return float, got {type(var)}"
        assert np.isfinite(var), f"VaR should be finite, got {var}"
        assert not np.isnan(var), f"VaR should not be NaN"
    
    @pytest.mark.parametrize("alpha", [0.1, 0.05, 0.01, 0.001])
    def test_es_at_different_levels(self, alpha):
        """Test ES calculation at different confidence levels."""
        es = alpha_stable.es(alpha, self.par)
        
        assert isinstance(es, float), f"ES should return float, got {type(es)}"
        assert np.isfinite(es), f"ES should be finite, got {es}"
        assert not np.isnan(es), f"ES should not be NaN"
    
    def test_var_es_consistency(self):
        """Test that VaR and ES follow risk measure properties."""
        for alpha in [0.05, 0.01]:
            var = alpha_stable.var(alpha, self.par)
            es = alpha_stable.es(alpha, self.par)
            check_var_es_properties(var, es, alpha)
    
    def test_var_alpha_validation(self):
        """Test that var validates alpha parameter."""
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            alpha_stable.var(1.5, self.par)
        
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            alpha_stable.var(-0.1, self.par)
    
    def test_es_alpha_validation(self):
        """Test that es validates alpha parameter."""
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            alpha_stable.es(1.5, self.par)
        
        with pytest.raises(ValueError, match="Alpha must be between 0 and 1"):
            alpha_stable.es(-0.1, self.par)
    
    def test_var_handles_scalar_return(self):
        """Test that var handles scalar return from ppf (regression test)."""
        # This was the bug: ppf returning scalar caused IndexError
        var = alpha_stable.var(0.05, self.par)
        assert isinstance(var, float)
        assert not np.isnan(var)


class TestAlphaStablePDF:
    """Test Alpha-Stable PDF functionality."""
    
    def test_pdf_scalar_input(self):
        """Test PDF with scalar input."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        result = alpha_stable.pdf(0.0, par)
        
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert result[0] > 0
        assert np.isfinite(result[0])
        assert not np.isnan(result[0])
    
    def test_pdf_array_input(self):
        """Test PDF with array input."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        x = np.array([-1.0, 0.0, 1.0])
        result = alpha_stable.pdf(x, par)
        
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all(result > 0)
        assert np.all(np.isfinite(result))
        assert np.all(np.logical_not(np.isnan(result)))
    
    def test_pdf_positive(self):
        """Test that PDF values are always positive."""
        returns = generate_test_returns(seed=42, size=100)
        par = alpha_stable.fit(returns)
        
        # Test at various points
        x = np.linspace(-0.1, 0.1, 10)
        result = alpha_stable.pdf(x, par)
        
        assert np.all(result > 0), "All PDF values should be positive"
        assert np.all(np.isfinite(result)), "All PDF values should be finite"
    
    def test_pdf_normal_special_case(self):
        """Test that alpha=2 gives normal distribution."""
        par = {'alpha': 2.0, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        
        # At alpha=2, should approximate normal
        result1 = alpha_stable.pdf(0.0, par)
        # Normal PDF at 0 with scale=1
        normal_pdf_0 = 1.0 / np.sqrt(2 * np.pi)
        assert np.isclose(result1[0], normal_pdf_0, rtol=0.1)


class TestAlphaStableCDF:
    """Test Alpha-Stable CDF functionality."""
    
    def test_cdf_scalar_input(self):
        """Test CDF with scalar input."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        result = alpha_stable.cdf(0.0, par)
        
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert 0 <= result[0] <= 1
    
    def test_cdf_array_input(self):
        """Test CDF with array input."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        x = np.array([-1.0, 0.0, 1.0])
        result = alpha_stable.cdf(x, par)
        
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all((result >= 0) & (result <= 1))
    
    def test_cdf_monotonic(self):
        """Test that CDF is monotonically increasing."""
        returns = generate_test_returns(seed=42, size=200)
        par = alpha_stable.fit(returns)
        
        x = np.linspace(-0.2, 0.2, 9)
        result = alpha_stable.cdf(x, par)
        
        assert np.all(np.diff(result) >= -1e-10), "CDF should be monotonically increasing"
    
    def test_cdf_boundaries(self):
        """Test CDF at extreme values."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        
        # Very negative should be near 0
        result_low = alpha_stable.cdf(-5.0, par)
        assert result_low[0] < 0.1, "CDF at very negative should be near 0"
        
        # Very positive should be near 1
        result_high = alpha_stable.cdf(5.0, par)
        assert result_high[0] > 0.9, "CDF at very positive should be near 1"


class TestAlphaStablePPF:
    """Test Alpha-Stable PPF functionality."""
    
    def test_ppf_scalar_input(self):
        """Test PPF with scalar input."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        result = alpha_stable.ppf(0.5, par)
        
        assert isinstance(result, np.ndarray)
        assert result.shape == (1,)
        assert np.isfinite(result[0])
        assert not np.isnan(result[0])
    
    def test_ppf_array_input(self):
        """Test PPF with array input."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        q = np.array([0.1, 0.5, 0.9])
        result = alpha_stable.ppf(q, par)
        
        assert isinstance(result, np.ndarray)
        assert result.shape == (3,)
        assert np.all(np.isfinite(result))
        assert np.all(np.logical_not(np.isnan(result)))
    
    @pytest.mark.parametrize("q", [0.05, 0.1, 0.5, 0.9, 0.95])
    def test_ppf_cdf_roundtrip(self, q):
        """Test that PPF and CDF are inverses."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        
        ppf_result = alpha_stable.ppf(q, par)
        cdf_result = alpha_stable.cdf(ppf_result, par)
        
        assert np.isclose(cdf_result[0], q, rtol=1e-4, atol=1e-4), \
            f"PPF->CDF roundtrip failed: expected {q}, got {cdf_result[0]}"
    
    def test_ppf_boundaries(self):
        """Test PPF at boundary values."""
        par = {'alpha': 1.8, 'beta': 0.0, 'c': 1.0, 'mu': 0.0}
        
        # Very low quantile should give very negative result
        result_low = alpha_stable.ppf(0.001, par)
        assert result_low[0] < -1.0
        
        # Very high quantile should give very positive result
        result_high = alpha_stable.ppf(0.999, par)
        assert result_high[0] > 1.0
